#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/FileSB.h>
#include <Xm/MessageB.h>
#include <X11/cursorfont.h>
#include <X11/Intrinsic.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <errno.h>
#include "SciPlot.h"
#include "global.h"

void barowin(char *title,
	     unsigned long nampli, unsigned long seqcount,
	     float brr[4][10000]) {
  Widget         toplevel, shell, main_w, menubar, f_menu;
  Widget         rowcol_w, plot_w, edit_w;
  Arg            args[5];
  int            i, graph1, graph2;
  unsigned  long n, nseq, ndata, ndown, nup;
  float          *skip;
  float          *sig, a, b, siga, sigb, chi2, q;
  float          slope, inters;
  char           *text, *fname;

  void file_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void fitul(float *x,float *y, unsigned long ndata, float *sig,
		    int mwt, float *a, float *b, float *siga,float *sigb,
		    float *chi2, float *q);

  /* initialize arrays */
  text=new char[1000];
  skip=new float[1];
  *skip=SCIPLOT_SKIP_VAL;

  toplevel=XtParent(drawing_a);
  while (!XtIsTopLevelShell(toplevel)) toplevel=XtParent(toplevel);
  shell=XtVaCreatePopupShell("barowin",
			     topLevelShellWidgetClass, toplevel,
			     XtNtitle, title,
			     NULL);
  main_w =XtVaCreateWidget("main_w", xmMainWindowWidgetClass, shell, NULL);
  menubar=XmVaCreateSimpleMenuBar(main_w, "menubar",
				  XmVaCASCADEBUTTON, XmStringCreateLocalized("File"), 'F',
				  NULL);
  f_menu=XmVaCreateSimplePulldownMenu(menubar, "file_menu", 0, file_cb,
			       XmVaPUSHBUTTON, XmStringCreateLocalized("Save"), 'S', NULL, NULL,
			       XmVaSEPARATOR,
			       XmVaPUSHBUTTON, XmStringCreateLocalized("Print"), 'P', NULL, NULL,
			       XmVaSEPARATOR,
			       XmVaPUSHBUTTON, XmStringCreateLocalized("Close"), 'C', NULL, NULL,
			       NULL);
  rowcol_w=XtVaCreateManagedWidget("rowcol", xmRowColumnWidgetClass, main_w,
  				   NULL);
  plot_w=XtVaCreateManagedWidget("plot", sciplotWidgetClass, rowcol_w,
				 XtNheight, 250,
				 XtNshowLegend, False,
				 XtNshowTitle, False,
				 XtNxLabel, "BP (mmHg)",
				 XtNshowYLabel, False,
				 XtNdrawMinor, True,
				 XtNyNumbersHorizontal, True,
				 XtNaxisFont, XtFONT_HELVETICA|12,
				 XtNlabelFont, XtFONT_HELVETICA|12,
				 XtNtitleFont, XtFONT_HELVETICA|12,
				 XtNmonochrome, True,
				 XtNchartType, XtCARTESIAN,
				 NULL);

  /* get xdat and ydat */
  nseq=0; ndata=0; ndown=0; nup=0; slope=0; inters=0;
  graph1=SciPlotListCreateFromFloat(plot_w, 0, skip, skip, "");
  graph2=SciPlotListCreateFromFloat(plot_w, 0, skip, skip, "");
  for (n=0; n<seqcount; n++) {
    if (brr[0][n]>0) {
      if (brr[1][ndata]<brr[1][ndata+1]) {
	nup++;
	SciPlotListAddFloat(plot_w, graph1, (int)brr[0][n], (brr[1]+ndata), (brr[2]+ndata));
	SciPlotListAddFloat(plot_w, graph1, 1, skip, skip);
	fitul((brr[1]+ndata-1), (brr[2]+ndata-1), (unsigned long)brr[0][n], sig, 0,
	      &a, &b, &siga, &sigb, &chi2, &q);
	slope+=b; inters+=a;
      }
      else {
	ndown++;
	SciPlotListAddFloat(plot_w, graph2, (int)brr[0][n], (brr[1]+ndata), (brr[2]+ndata));
	SciPlotListAddFloat(plot_w, graph2, 1, skip, skip);
	fitul((brr[1]+ndata-1), (brr[2]+ndata-1), (unsigned long)brr[0][n], sig, 0,
	      &a, &b, &siga, &sigb, &chi2, &q);
	slope+=b; inters+=a;
      }
      ndata+=(int)brr[0][n];
      nseq++;
    }
    else ndata-=(int)brr[0][n];
  }
  if (nseq>0) {
    slope=slope/nseq; inters=inters/nseq;
  }
  SciPlotListSetStyle(plot_w, graph1, 1, XtMARKER_FCIRCLE, 1, XtLINE_SOLID);
  SciPlotListSetStyle(plot_w, graph2, 1, XtMARKER_CIRCLE, 1, XtLINE_SOLID);
  SciPlotSetXAutoScale(plot_w);
  SciPlotSetYAutoScale(plot_w);

  i=0;
  XtSetArg(args[i], XmNrows, 8); i++;
  XtSetArg(args[i], XmNcolumns, 80); i++;
  XtSetArg(args[i], XmNeditMode, XmMULTI_LINE_EDIT); i++;
  XtSetArg(args[i], XmNuserData, plot_w); i++;
  edit_w=XmCreateText(rowcol_w, "edit_w", args, i);
  /* get current file name */
  toplevel=XtParent(drawing_a);
  while (!XtIsTopLevelShell(toplevel)) toplevel=XtParent(toplevel);
  XtVaGetValues(toplevel, XmNtitle, &fname, NULL);
  fname=strrchr(fname, ' ')+1;
  /* write report in edit_w */
  XmTextSetString(edit_w, "Analysis of baroreflex according to Bertinieri \n");
  sprintf(text, "Data file                : %s \n", fname);
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Number of BP amplitudes  : %d \n", nampli); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Number of total sequences: %d \n", nseq); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Number of up-sequences   : %d (filled circles) \n", nup); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Number of down-sequences : %d (open circles)\n", ndown); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Mean sensitivity (slope) : %f \n", slope); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);
  sprintf(text, "Mean y-axis intersept    : %f \n", inters); 
  XmTextInsert(edit_w, XmTextGetLastPosition(edit_w), text);

  XtManageChild(edit_w);
  XtManageChild(menubar);
  XtManageChild(main_w);
  XtVaSetValues(f_menu, XmNuserData, edit_w, NULL);

  XtPopup(shell, XtGrabNone);
  SciPlotUpdate(plot_w);

  delete[]text;
}

/*   .................   */
/*   c a l l b a c k s   */
/*   .................   */

void file_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int    selection;
  Widget toplevel, fsb_sd, edit_w, plot_w, dialog;
  Arg    args[5];
  char   cmd[255];
  int    i;

  void baro_print_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void save_results_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void cancel_cb(Widget, XtPointer, XtPointer);
  extern void printhelp_cb(Widget, XtPointer, XtPointer);
  
  XtVaGetValues(XtParent(widget), XmNuserData, &edit_w, NULL);
  XtVaGetValues(edit_w, XmNuserData, &plot_w, NULL);
  selection=(int) client_data;

  switch (selection) {
  case 0: /* save a file */;
    toplevel=XtParent(widget);
    while (!XtIsTopLevelShell(toplevel)) toplevel=XtParent(toplevel);
    fsb_sd=XmCreateFileSelectionDialog(toplevel, "filesb", NULL, 0);
    XtAddCallback(fsb_sd,XmNokCallback, save_results_cb, NULL);
    XtAddCallback(fsb_sd,XmNcancelCallback, cancel_cb, NULL);
    XtVaSetValues(XtParent(fsb_sd),
		  XmNtitle, "Save Baroreflex Data",
		  NULL);
    XtVaSetValues(fsb_sd, XmNuserData, plot_w, NULL);
    XtManageChild(fsb_sd);
    break;
  case 1: /* print a file */;
    i=0;
    XtSetArg(args[i], XmNselectionLabelString, XmStringCreateLocalized("Enter Print Command: ")); i++;
    XtSetArg(args[i], XmNtitle, "Printing"); i++;
    dialog=XmCreatePromptDialog(plot_w, "dialog", args, i);
    strcpy(cmd, PRINTCMD);
    XtVaSetValues(dialog,
		  XmNtextString, XmStringCreateLocalized(cmd),
		  NULL);
    XtAddCallback(dialog, XmNokCallback, baro_print_cb, edit_w);
    XtAddCallback(dialog, XmNhelpCallback, printhelp_cb, NULL);

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
    break;
  case 2: /* close editor */;
    while(widget && !XtIsWMShell(widget)) widget=XtParent(widget);
    XtPopdown(widget);
    XtDestroyWidget(widget);
    break;
  }
}

/* print callback */
void baro_print_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Widget plot_w, edit_w;
  FILE   *pdat;
  XmSelectionBoxCallbackStruct *cbs=(XmSelectionBoxCallbackStruct *) call_data;
  char   *text, *ptext, cmd[255];
  long   len;
  int    i, ypos, linewidth;

  /* build temporary print file */
  edit_w=(Widget)client_data;
  XtVaGetValues(edit_w, XmNuserData, &plot_w, NULL);
  SciPlotPSCreate(plot_w, PRINTDAT, 3, 0);
  text=XmTextGetString(edit_w);

  pdat=fopen(PRINTDAT, "a");
  strcpy(cmd, "% print report \n");
  fwrite(cmd, sizeof(char), strlen(cmd), pdat);
  strcpy(cmd, "/report-font {/Courier findfont 12 scalefont setfont} bind def\n");
  fwrite(cmd, sizeof(char), strlen(cmd), pdat);

  ptext=text; ypos=450; linewidth=15; i=0;
  while (*ptext) {
    sprintf(cmd, "report-font 50.0 %d ma (", ypos-i*linewidth);
    len=strcspn(ptext,"\n\0");
    strncat(cmd, ptext, len);
    strcat(cmd, ") show\n");
    fwrite(cmd, sizeof(char), strlen(cmd), pdat);
    ptext+=len;
    if (*ptext=='\n') ptext++;
    i++;
  }
  strcpy(cmd, "showpage\n");
  fwrite(cmd, sizeof(char), strlen(cmd), pdat);
  fclose(pdat);
  XtFree(text);
  /* get print command and print */
  XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &text);
  strcpy(cmd, text); strcat(cmd, " "); strcat(cmd, PRINTDAT);
  system(cmd);
  XtFree(text);
}


/* save results callback */
void save_results_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Widget       plot_w, fexist;
  Cursor       uhr, pfeil;
  Arg          args[5];
  int          i;
  XmString     fname;
  char         *filename, *out_buf;
  ofstream     outdat;
  FILE         *datei;
  //  extern int   errno;

  void kill_resfile_cb(Widget widget, XtPointer client_data, XtPointer call_data);

  XtVaGetValues(widget, XmNuserData, &plot_w, NULL);

  /* get filename */
  XtVaGetValues(widget, XmNdirSpec, &fname, NULL);
  if (!XmStringGetLtoR(fname, XmFONTLIST_DEFAULT_TAG, &filename)) {
    cout << "Internal error pasting filename! \n";
    return;
  }
  if (!*filename) {
    cout << "No file selected! \n";
    XtFree(filename);
    return;
  }

  /* test whether file already exists */
  outdat.open(filename, ios::noreplace);
  if (errno==EEXIST) {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("File already exists.\n     Overwrite?",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNokLabelString, XmStringCreateLocalized(" Yes ")); i++;
    XtSetArg(args[i], XmNcancelLabelString, XmStringCreateLocalized(" No ")); i++;
    XtSetArg(args[i], XmNtitle, "Save Results"); i++;
    XtSetArg(args[i], XmNuserData, plot_w); i++;
    fexist=XmCreateQuestionDialog(XtParent(widget), "fexist", args, i);
    XtAddCallback(fexist, XmNokCallback, kill_resfile_cb, widget);
    XtUnmanageChild(XmMessageBoxGetChild(fexist, XmDIALOG_HELP_BUTTON));
    XtManageChild(fexist);
    XtPopup(XtParent(fexist), XtGrabNone);
  }

  /* file does not exist - now write data */
  if (outdat) {
    outdat.close();
    /* set cursor */
    uhr  =XCreateFontCursor(XtDisplay(widget),XC_watch);
    pfeil=XCreateFontCursor(XtDisplay(widget), XC_arrow);
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    // write file;
    if ((datei=fopen(filename, "w"))==NULL) {
      cout << "Can't open file: " << filename << "\n";
    }
    else {
      SciPlotExportData(plot_w, datei); 
      fclose(datei);
    }
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    XFreeCursor(XtDisplay(widget), pfeil);
    XFreeCursor(XtDisplay(widget), uhr);
    XtUnmanageChild(widget);
  }
  XtFree(filename);

}


void kill_resfile_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Widget   fsb, plot_w;
  XmString fname;
  char     *filename, *cmd;

  cmd=new char[256];
  XtVaGetValues(widget, XmNuserData, &plot_w, NULL);
  fsb=(Widget) client_data;
  XtVaGetValues(fsb, XmNdirSpec, &fname, NULL);
  if (XmStringGetLtoR(fname, XmFONTLIST_DEFAULT_TAG, &filename)) {
    sprintf(cmd, "rm %s", filename);
    system(cmd);
    XtFree(filename);
    save_results_cb(fsb, (XtPointer)plot_w, NULL);
  }
  delete[]cmd;
}
