#include <Xm/Xm.h>
#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleBG.h>
#include <Xm/Scale.h>
#include <Xm/LabelG.h>
#include <Xm/TextF.h>
#include <X11/cursorfont.h>
#include <stdlib.h>
#include <math.h>
#include "global.h"
#include "nrc/nrc.h"

typedef struct {
  int   type;    // 0: low pass, 1: high pass;
  int   order;   // order for butterworth filter;
  float freq;    // corner frequency;
} filter_struct;

typedef struct {
  float m; // m in y = m*x +b;
  float b; // b in y = m*x +b;
} linear_struct;

typedef struct {
  float tau;      // time in seconds to average;
  short reduce;   // reduce sr (1) or not (0);
  short is_error; // 0: no error, !0: error;
} mavg_struct;

typedef struct {
  float vor, nach; // time before and after trigger;
  float schwelle;  // threshold for trigger;
} savg_struct;

typedef struct {
  Widget rb[6];
  short  rb_selected;
  short  ch1, ch2;
} twochtrans_struct;

void c_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
          int           item;
          int           i, nactive, chn, chn1, chn2, lastch;
          unsigned long fn, n, na, ne;
          Widget        mbox, rowcol, frame, toggle_box, low_tg, high_tg, rsr_tg, w;
          Arg           args[5];
	  Cursor        uhr, pfeil;
          char          text[255];
	  float         ave, adev, sdev, var, skew, curt;
	  float         *fft_data;
	  fcomplex      z;
  static  float         interpol_sr=1000.0;
  static  Widget        tb[10];
  static  filter_struct *filter;
  static  linear_struct *linear;
  static  baro_struct   *baro;
  static  mavg_struct   *mavg;
  static  savg_struct   *savg;
  static  twochtrans_struct *twochtrans;

  void cancel_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void derived_c_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void derived_c_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void derived_t_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void derived_t_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void filter_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void filter_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void filter_tg_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void filter_ord_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void filter_freq_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void linear_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void linear_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void linear_pm_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void linear_pb_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void interpol_sr_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void interpol_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void interpol_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void baro_ch_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void baro_tb_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void baro_thre_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void baro_reg_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void baro_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void mavg_n_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void mavg_tg_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void mavg_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void mavg_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void seqavg_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void seqavg_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void seqavg_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void twochtrans_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void twochtrans_help_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void baro_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void showmess(Widget widget, char *title, char *text);

  if (!filter) {
    filter=new filter_struct;
    filter->type=0;
    filter->order=4;
    filter->freq=3.0; 
  }

  if (!linear) {
    linear=new linear_struct;
    linear->m=1;
    linear->b=0;
  }

  if (!baro) {
    baro=new baro_struct;
    baro->bp_ch=1;
    baro->smd=1;
    baro->ch=0;
    baro->thre=20;
    baro->reg=0.80;
  }

  if (!mavg) {
    mavg=new mavg_struct;
    mavg->tau=1.0;
    mavg->reduce=0;
    mavg->is_error=0;
  }

  if (!savg) {
    savg=new savg_struct;
    savg->vor=0.2;
    savg->nach=0.5;
    savg->schwelle=20000;
  }

  if (!twochtrans) {
    twochtrans=new twochtrans_struct;
    twochtrans->rb_selected=0;
    twochtrans->ch1=-1;
    twochtrans->ch2=-1;
  }

  uhr  =XCreateFontCursor(XtDisplay(drawing_a),XC_watch);
  pfeil=XCreateFontCursor(XtDisplay(drawing_a), XC_arrow);

  item=(int) client_data;
  switch(item) {
  case 0:  // heart rate by Threshold;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive==1) {
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Derived channels by threshold"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNselectionLabelString, XmStringCreateLocalized("Threshold:")); i++;
      mbox=XmCreatePromptDialog(drawing_a, "derived", args, i);
      // create rowcolumn widget;
      toggle_box=XtVaCreateWidget("togglebox",
				  xmRowColumnWidgetClass, mbox,
				  XmNpacking, XmPACK_COLUMN,
				  XmNnumColumns, 2,
				  XmNradioBehavior, False,
				  NULL);
      // create and set toggles;
      tb[0]=XtVaCreateManagedWidget("frequency", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[1]=XtVaCreateManagedWidget("maximum", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[2]=XtVaCreateManagedWidget("mean", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[3]=XtVaCreateManagedWidget("minimum", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[4]=XtVaCreateManagedWidget("amplitude", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[5]=XtVaCreateManagedWidget("dy/dt", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[6]=XtVaCreateManagedWidget("dy/dt_max", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[7]=XtVaCreateManagedWidget("dy/dt_min", xmToggleButtonGadgetClass, toggle_box, NULL);
      XmToggleButtonGadgetSetState(tb[0],True,False);
      XmToggleButtonGadgetSetState(tb[1],False,False);
      XmToggleButtonGadgetSetState(tb[2],False,False);
      XmToggleButtonGadgetSetState(tb[3],False,False);
      XmToggleButtonGadgetSetState(tb[4],False,False);
      XmToggleButtonGadgetSetState(tb[5],False,False);
      XmToggleButtonGadgetSetState(tb[6],False,False);
      XmToggleButtonGadgetSetState(tb[7],False,False);
      XtAddCallback(mbox, XmNokCallback, derived_t_ok_cb, tb);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, derived_t_help_cb, NULL);
      XtManageChild(toggle_box);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    }
    else {
      showmess(drawing_a, "Derived channels by threshold", "Exactly 1 channel must be active!");
    }
    break;
  case 1:      // derived channels by convolution;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive==1) {
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Derived channels by convolution"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      mbox=XmCreateMessageDialog(drawing_a, "derived", args, i);
      // create rowcolumn widget;
      toggle_box=XtVaCreateWidget("togglebox",
				  xmRowColumnWidgetClass, mbox,
				  XmNpacking, XmPACK_COLUMN,
				  XmNnumColumns, 2,
				  XmNradioBehavior, False,
				  NULL);
      // create and set toggles;
      tb[0]=XtVaCreateManagedWidget("frequency", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[1]=XtVaCreateManagedWidget("maximum", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[2]=XtVaCreateManagedWidget("mean", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[3]=XtVaCreateManagedWidget("minimum", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[4]=XtVaCreateManagedWidget("amplitude", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[5]=XtVaCreateManagedWidget("dy/dt", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[6]=XtVaCreateManagedWidget("dy/dt_max", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[7]=XtVaCreateManagedWidget("dy/dt_min", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[8]=XtVaCreateManagedWidget("trigger curve", xmToggleButtonGadgetClass, toggle_box, NULL);
      tb[9]=XtVaCreateManagedWidget("trigger value", xmToggleButtonGadgetClass, toggle_box, NULL);
      XmToggleButtonGadgetSetState(tb[0],True,False);
      XmToggleButtonGadgetSetState(tb[1],False,False);
      XmToggleButtonGadgetSetState(tb[2],False,False);
      XmToggleButtonGadgetSetState(tb[3],False,False);
      XmToggleButtonGadgetSetState(tb[4],False,False);
      XmToggleButtonGadgetSetState(tb[5],False,False);
      XmToggleButtonGadgetSetState(tb[6],False,False);
      XmToggleButtonGadgetSetState(tb[7],False,False);
      XmToggleButtonGadgetSetState(tb[8],False,False);
      XmToggleButtonGadgetSetState(tb[9],False,False);
      XtAddCallback(mbox, XmNokCallback, derived_c_ok_cb, tb);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, derived_c_help_cb, NULL);
      XtManageChild(toggle_box);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    }
    else {
      showmess(drawing_a, "Derived channels by convolution", "Exactly 1 channel must be active!");
    }
    break;
  case 2:   // butterworth filter;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive>0) {
      /* determine 2^n for FFT */
      fn=(unsigned long)pow(2,ceil(log(r.ch[chn].n)/log(2))); 
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Butterworth Filter"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString, XmStringCreateLocalized("Parameters for Butterworth Filter: ")); i++;
      mbox=XmCreateMessageDialog(drawing_a, "filter", args, i);
      // create rowcolumn widget;
      rowcol=XtVaCreateWidget("rowcol",
			      xmRowColumnWidgetClass, mbox,
			      NULL);
      sprintf(text, "(FFT based on %u values)", fn);
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      // create and set togglebox;
      frame=XtVaCreateWidget("frame", xmFrameWidgetClass, rowcol, NULL);
      toggle_box=XmCreateRadioBox(frame, "radio_box", NULL, 0);
      low_tg=XtVaCreateManagedWidget("Low pass", xmToggleButtonGadgetClass, toggle_box,
				     XmNuserData, filter,
				     NULL);
      XtAddCallback(low_tg, XmNvalueChangedCallback, filter_tg_cb, (XtPointer)0);
      high_tg=XtVaCreateManagedWidget("High pass", xmToggleButtonGadgetClass, toggle_box,
				      XmNuserData, filter,
				      NULL);
      XtAddCallback(high_tg, XmNvalueChangedCallback, filter_tg_cb, (XtPointer)1);
      if (filter->type==0) {
	XmToggleButtonGadgetSetState(low_tg,True,False);
	XmToggleButtonGadgetSetState(high_tg,False,False);
      }
      if (filter->type==1) {
	XmToggleButtonGadgetSetState(low_tg,False,False);
	XmToggleButtonGadgetSetState(high_tg,True,False);
      }
      w=XtVaCreateManagedWidget("fiorder", xmScaleWidgetClass, rowcol,
			      XtVaTypedArg, XmNtitleString, XmRString, "Filter Order", 13,
			      XmNorientation, XmHORIZONTAL,
			      XmNmaximum, 10,
			      XmNminimum, 1,
			      XmNvalue, filter->order,
			      XmNshowValue, True,
			      NULL);
      XtAddCallback(w, XmNvalueChangedCallback, filter_ord_cb, filter);
      XtVaCreateManagedWidget("Corner Frequency:", xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text, "%f", filter->freq);
      w=XtVaCreateManagedWidget("fifreq", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, filter_freq_cb, filter);
      XtAddCallback(mbox, XmNokCallback, filter_ok_cb, filter);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, filter_help_cb, NULL); 
      XtManageChild(toggle_box);
      XtManageChild(frame);
      XtManageChild(rowcol);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    }
    else {
      showmess(drawing_a, "Butterworth Filter", "At least 1 channel must be active!");
    }
    break;
  case 3: // interpolate;
    nactive=0;
    for (i=0; i<r.n; i++)
      if (r.view[i].is_active && r.view[i].is_visible) nactive++;

    if (r.nvis>0 && nactive>0) {
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Spline Interpolation"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString, XmStringCreateLocalized("new sampling rate: ")); i++;
      mbox=XmCreateMessageDialog(drawing_a, "interpol", args, i);
      sprintf(text, "%f", interpol_sr);
      w=XtVaCreateManagedWidget("interpol_sr", xmTextFieldWidgetClass, mbox,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, interpol_sr_cb, &interpol_sr);
      XtAddCallback(mbox, XmNokCallback, interpol_ok_cb, &interpol_sr);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, interpol_help_cb, NULL);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    } /* endif */
    else {
      showmess(drawing_a, "Spline Interpolation", "At least 1 channel must be active!");
    }
    break;
  case 4: // moving average;
    /* get number of active channels and first active channel */
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }
    /* if at least one channel is active -> do something */
    if (r.nvis>0 && nactive>0) {
      i=0;
      XtSetArg(args[i], XmNtitle, "Moving Average"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString, XmStringCreateLocalized("Moving Average: ")); i++;
      mbox=XmCreateMessageDialog(drawing_a, "movavg", args, i);
      // create rowcolumn widget;
      rowcol=XtVaCreateWidget("rowcol",
			      xmRowColumnWidgetClass, mbox,
			      NULL);
      XtVaCreateManagedWidget("time in seconds to average:", xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text, "%f", mavg->tau);
      w=XtVaCreateManagedWidget("tmovavg", xmTextFieldWidgetClass, rowcol,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, mavg_n_cb, mavg);
      // create and set togglebox;
      frame=XtVaCreateWidget("frame", xmFrameWidgetClass, rowcol, NULL);
      rsr_tg=XtVaCreateManagedWidget("reduce sampling rate", xmToggleButtonGadgetClass, frame,
				     NULL);
      XtAddCallback(rsr_tg, XmNvalueChangedCallback, mavg_tg_cb, mavg);
      if (mavg->reduce==0) {
	XmToggleButtonGadgetSetState(rsr_tg,False,False);
      }
      if (mavg->reduce==1) {
	XmToggleButtonGadgetSetState(rsr_tg,True,False);
      }
      XtAddCallback(mbox, XmNokCallback, mavg_ok_cb, mavg);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, mavg_help_cb, NULL);
      XtManageChild(frame);
      XtManageChild(rowcol);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    }
    /* no channel is active */
    else {
      showmess(drawing_a, "Moving Average", "At least 1 channel must be active!");
    }
    
    break;
  case 5:  // rectify;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }
    if (r.nvis>0 && nactive>0) {
      for (i=0; i<r.n; i++) {
	if (r.view[i].is_active && r.view[i].is_visible) {
	  moment(r.ch[i].data-1,r.ch[i].n,&ave,&adev,&sdev,&var,&skew,&curt);
	  r.ch[i].min=fabs(r.ch[i].data[0]-ave);
	  r.ch[i].max=fabs(r.ch[i].data[0]-ave);
	  for (n=0; n<r.ch[i].n; n++) {
	    r.ch[i].data[n]=fabs(r.ch[i].data[n]-ave);
	    if (r.ch[i].min > r.ch[i].data[n])
	      r.ch[i].min=r.ch[i].data[n];
	    if (r.ch[i].max < r.ch[i].data[n])
	      r.ch[i].max=r.ch[i].data[n];
	  } /* endfor */

	  na=(unsigned long)(r.t_1*r.ch[i].sr);
	  if (na>=r.ch[i].n) na=r.ch[i].n-1;
	  ne=(unsigned long)(r.t_2*r.ch[i].sr);
	  if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	  r.view[i].min=1e30; r.view[i].max=-1e30;
	  for (n=na; n<=ne; n++) {
	    if (r.view[i].min > r.ch[i].data[n])
	      r.view[i].min=r.ch[i].data[n];
	    if (r.view[i].max < r.ch[i].data[n])
	      r.view[i].max=r.ch[i].data[n];
	  }
	} /* endif */
      } /* endfor */
      da_cb(NULL, NULL, NULL);
    }
    else {
      showmess(drawing_a, "Rectify", "At least one channel must be active!");
    }
    break;
  case 6:  // invert;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive>0) {
      for (i=0; i<r.n; i++) {
	if (r.view[i].is_active && r.view[i].is_visible) {
	  moment(r.ch[i].data-1,r.ch[i].n,&ave,&adev,&sdev,&var,&skew,&curt);
	  r.ch[i].min=(-1)*(r.ch[i].data[0]-ave)+ave;
	  r.ch[i].max=(-1)*(r.ch[i].data[0]-ave)+ave;
	  for (n=0; n<r.ch[i].n; n++) {
	    r.ch[i].data[n]=(-1)*(r.ch[i].data[n]-ave)+ave;
	    if (r.ch[i].min > r.ch[i].data[n])
	      r.ch[i].min=r.ch[i].data[n];
	    if (r.ch[i].max < r.ch[i].data[n])
	      r.ch[i].max=r.ch[i].data[n];
	  } /* endfor */

	  na=(unsigned long)(r.t_1*r.ch[i].sr);
	  if (na>=r.ch[i].n) na=r.ch[i].n-1;
	  ne=(unsigned long)(r.t_2*r.ch[i].sr);
	  if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	  r.view[i].min=1e30; r.view[i].max=-1e30;
	  for (n=na; n<=ne; n++) {
	    if (r.view[i].min > r.ch[i].data[n])
	      r.view[i].min=r.ch[i].data[n];
	    if (r.view[i].max < r.ch[i].data[n])
	      r.view[i].max=r.ch[i].data[n];
	  }
	} /* endif */
      } /* endfor */
      da_cb(NULL, NULL, NULL);
    }
    else {
      showmess(drawing_a, "Invert", "At least one channel must be active!");
    }
    break;
  case 7:  // linear transformation;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive>0) {
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Linear Transformation"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString,
	       XmStringCreateLtoR("linear transformation: y = m * x + b ",
				  XmFONTLIST_DEFAULT_TAG)); i++;
      mbox=XmCreateMessageDialog(drawing_a, "linear", args, i);
      // create rowcolumn widget;
      rowcol=XtVaCreateWidget("rowcol",
			      xmRowColumnWidgetClass, mbox,
			      XmNorientation, XmHORIZONTAL,
			      XmNnumColumns, 2,
			      XmNpacking, XmPACK_COLUMN,
			      NULL);
      strcpy(text, "parameter m: ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",linear->m);
      w=XtVaCreateManagedWidget("para_m", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, linear_pm_cb, linear);
      strcpy(text, "parameter b: ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",linear->b);
      w=XtVaCreateManagedWidget("para_b", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, linear_pb_cb, linear);
      XtAddCallback(mbox, XmNokCallback, linear_ok_cb, linear);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, linear_help_cb, NULL);
      XtManageChild(rowcol);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    } /* endif */
    else {
      showmess(drawing_a, "Linear Transformation", "At least one channel must be active!");
    }
    break;
  case 8:   /* Fast Fourier Transform */
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }
    /* and here it goes ... */
    if (r.nvis>0 && nactive>0) {
      XDefineCursor(XtDisplay(drawing_a),XtWindow(drawing_a),uhr);
      XFlush(XtDisplay(drawing_a));
      lastch=r.n;
      for (i=0; i<lastch; i++) {
	if (r.view[i].is_active && r.view[i].is_visible) {
	  /* determine na and ne */
	  na=(unsigned long)(r.ts_1*r.ch[i].sr);
	  if (na>=r.ch[i].n) na=r.ch[i].n-1;
	  ne=(unsigned long)(r.ts_2*r.ch[i].sr);
	  if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	  if (ne>na) {
	    /* determine 2^n for FFT */
	    fn=(unsigned long)pow(2,ceil(log(ne-na+1)/log(2))); 
	    /* Dataarray Initialization */
	    fft_data=new float[2*fn*sizeof(float)];
	    /* copy data in fft_data */
	    for (n=na; n<=ne; n++) fft_data[n-na]=r.ch[i].data[n];
	    /* mean for zero padding */
	    moment(fft_data-1,(ne-na+1),&ave,&adev,&sdev,&var,&skew,&curt);
	    /* Daten kopieren */
	    for (n=na; n<=ne; n++) {
	      fft_data[2*(n-na)]=r.ch[i].data[n]-ave;
	      fft_data[2*(n-na)+1]=0;
	    } /* endfor */
	    for (n=ne-na+1; n<fn; n++) {
	      fft_data[2*n]=0;
	      fft_data[2*n+1]=0;
	    } /* endfor */
	    /* fast fourier transform */
	    four1(fft_data-1,fn,1);
	    r.AddChannel(fn/2+1, r.ch[i].sr, r.ch[i].min, r.ch[i].max,
			 r.ch[i].height, r.ch[i].color,
			 r.ch[i].name, r.ch[i].unit);
	    r.view[r.n-1].is_frequency=1;
	    // in neue REGIST copieren;
	    z.r=fft_data[0];z.i=fft_data[1];
	    r.ch[r.n-1].data[0]=(Cabs(z)*Cabs(z)/fn/fn)*fn/r.ch[i].sr;
	    for (n=1; n<fn/2;n++) {
	      z.r=fft_data[n*2];z.i=fft_data[n*2+1];
	      r.ch[r.n-1].data[n]=(2*Cabs(z)*Cabs(z)/fn/fn)*fn/r.ch[i].sr;
	    } /* endfor */
	    z.r=fft_data[fn];z.i=fft_data[fn+1];
	    r.ch[r.n-1].data[fn/2]=(Cabs(z)*Cabs(z)/fn/fn)*fn/r.ch[i].sr;
	    /* Min und Max fuer Kanal setzen */
	    na=(unsigned long)(r.t_1*r.ch[r.n-1].sr);
	    if (na>=r.ch[r.n-1].n) na=r.ch[r.n-1].n-1;
	    ne=(unsigned long)(r.t_2*r.ch[r.n-1].sr);
	    if (ne>=r.ch[r.n-1].n) ne=r.ch[r.n-1].n-1;
	    r.view[r.n-1].min=1e30; r.view[r.n-1].max=-1e30;
	    for (n=na; n<=ne; n++) {
	      if (r.view[r.n-1].min > r.ch[r.n-1].data[n])
		r.view[r.n-1].min=r.ch[r.n-1].data[n];
	      if (r.view[r.n-1].max < r.ch[r.n-1].data[n])
		r.view[r.n-1].max=r.ch[r.n-1].data[n];
	    }
	    r.ch[r.n-1].min=1e30;   r.ch[r.n-1].max=-1e30;
	    for (n=0; n<r.ch[r.n-1].n; n++) {
	      if (r.ch[r.n-1].data[n]<r.ch[r.n-1].min) r.ch[r.n-1].min=r.ch[r.n-1].data[n];
	      if (r.ch[r.n-1].data[n]>r.ch[r.n-1].max) r.ch[r.n-1].max=r.ch[r.n-1].data[n];
	    } /* endfor */
	    /* End of calculation */
	    delete[]fft_data;
	  }
	}
      }
      XDefineCursor(XtDisplay(drawing_a),XtWindow(drawing_a),pfeil);
      da_cb(NULL, NULL, NULL);
    }
    else {
      showmess(drawing_a, "Fast Fourier Transformation", "At least one channel must be active!");
    }
    break;
  case 9: // baroreflex according to Bertinieri;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive==1) {
      baro->bp_ch=chn;
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Baroreflex"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString,
	       XmStringCreateLtoR("Baroreflex according to Bertinieri",
				  XmFONTLIST_DEFAULT_TAG)); i++;
      mbox=XmCreateMessageDialog(drawing_a, "baro", args, i);
      // create rowcolumn widget;
      rowcol=XtVaCreateWidget("rowcol",
			      xmRowColumnWidgetClass, mbox,
			      XmNorientation, XmHORIZONTAL,
			      XmNnumColumns, 4,
			      XmNpacking, XmPACK_COLUMN,
			      NULL);
      strcpy(text, "channel for y-axis: ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%d",baro->ch);
      w=XtVaCreateManagedWidget("baro_ch", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, baro_ch_cb, baro);
      strcpy(text, "Threshold: ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text, "%f", baro->thre);
      w=XtVaCreateManagedWidget("baro_thre", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, baro_thre_cb, baro);
      strcpy(text, "r for inclusion: ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",baro->reg);
      w=XtVaCreateManagedWidget("baro_reg", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      strcpy(text, "use:");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      /* toggle buttons for SYS, MEA, DIA */
      toggle_box=XtVaCreateWidget("togglebox",
				  xmRowColumnWidgetClass, rowcol,
				  XmNorientation, XmHORIZONTAL,
				  XmNnumColumns, 1,
				  XmNpacking, XmPACK_COLUMN,
				  XmNradioBehavior, True,
				  NULL);
      // create and set toggles;
      tb[0]=XtVaCreateManagedWidget("SYS", xmToggleButtonGadgetClass, toggle_box,
				    XmNuserData, 1,
				    NULL);
      XtAddCallback(tb[0], XmNvalueChangedCallback, baro_tb_cb, baro);
      tb[1]=XtVaCreateManagedWidget("Mean", xmToggleButtonGadgetClass, toggle_box,
				    XmNuserData, 2,
				    NULL);
      XtAddCallback(tb[1], XmNvalueChangedCallback, baro_tb_cb, baro);
      tb[2]=XtVaCreateManagedWidget("DIA", xmToggleButtonGadgetClass, toggle_box,
				    XmNuserData, 3,
				    NULL);
      XtAddCallback(tb[2], XmNvalueChangedCallback, baro_tb_cb, baro);
      if (baro->smd==1) XmToggleButtonGadgetSetState(tb[0],True,False);
      else if (baro->smd==2) XmToggleButtonGadgetSetState(tb[1],True,False);
      else if (baro->smd==3) XmToggleButtonGadgetSetState(tb[2],True,False);
      else XmToggleButtonGadgetSetState(tb[0],True,False);
      /* end of toggles */      
      XtAddCallback(w, XmNvalueChangedCallback, baro_reg_cb, baro);
      XtAddCallback(mbox, XmNokCallback, baro_ok_cb, baro);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, baro_help_cb, NULL);
      XtManageChild(toggle_box);
      XtManageChild(rowcol);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    } /* endif */
    else {
      showmess(drawing_a,
	       "Baroreflex",
	       "The blood pressure channel must be active!");
    }
    break;
  case 10: // differentiate;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }

    if (r.nvis>0 && nactive>0) {
      for (i=0; i<r.n; i++) {
	if (r.view[i].is_active && r.view[i].is_visible) {
	  r.ch[i].min=0;
	  r.ch[i].max=0;
	  for (n=0; n<r.ch[i].n-1; n++) {
	    r.ch[i].data[n]=(r.ch[i].data[n+1]-r.ch[i].data[n])*r.ch[i].sr;
	    if (r.ch[i].min > r.ch[i].data[n])
	      r.ch[i].min=r.ch[i].data[n];
	    if (r.ch[i].max < r.ch[i].data[n])
	      r.ch[i].max=r.ch[i].data[n];
	  } /* endfor */
	  r.ch[i].data[r.ch[i].n-1]=r.ch[i].data[r.ch[i].n-2];

	  na=(unsigned long)(r.t_1*r.ch[i].sr);
	  if (na>=r.ch[i].n) na=r.ch[i].n-1;
	  ne=(unsigned long)(r.t_2*r.ch[i].sr);
	  if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	  r.view[i].min=1e30; r.view[i].max=-1e30;
	  for (n=na; n<=ne; n++) {
	    if (r.view[i].min > r.ch[i].data[n])
	      r.view[i].min=r.ch[i].data[n];
	    if (r.view[i].max < r.ch[i].data[n])
	      r.view[i].max=r.ch[i].data[n];
	  } /* endfor */
	} /* endif */
      } /* endfor */
      da_cb(NULL, NULL, NULL);
    }
    else {
      showmess(drawing_a, "Differentiate", "At least 1 channel must be active!");
    }
    break;
  case 11:  // sequential average;
    nactive=0; chn=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (chn<0) chn=i;
      }
    }
    // at least one channel is active;
    if (r.nvis>0 && nactive>0) {
      // create dialog box;
      i=0;
      XtSetArg(args[i], XmNtitle, "Sequential Average"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString,
	       XmStringCreateLtoR("Window size and threshold:",
				  XmFONTLIST_DEFAULT_TAG)); i++;
      mbox=XmCreateMessageDialog(drawing_a, "sequential", args, i);
      // create rowcolumn widget;
      rowcol=XtVaCreateWidget("rowcol",
			      xmRowColumnWidgetClass, mbox,
			      XmNorientation, XmHORIZONTAL,
			      XmNnumColumns, 3,
			      XmNpacking, XmPACK_COLUMN,
			      NULL);
      strcpy(text, "Time before trigger (s): ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",savg->vor);
      w=XtVaCreateManagedWidget("seqavgv", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, seqavg_cb, &savg->vor);
      strcpy(text, "Time after  trigger (s): ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",savg->nach);
      w=XtVaCreateManagedWidget("seqavgn", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, seqavg_cb, &savg->nach);
      strcpy(text, "Threshold              : ");
      XtVaCreateManagedWidget(text, xmLabelGadgetClass, rowcol,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
      sprintf(text,"%f",savg->schwelle);
      w=XtVaCreateManagedWidget("seqavgs", xmTextFieldWidgetClass, rowcol,
				XmNeditMode, XmSINGLE_LINE_EDIT,
				XmNvalue, text,
				NULL);
      XtAddCallback(w, XmNvalueChangedCallback, seqavg_cb, &savg->schwelle);
      XtAddCallback(mbox, XmNokCallback, seqavg_ok_cb, savg);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, seqavg_help_cb, NULL);
      XtManageChild(rowcol);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    } /* endif */
    else {
      showmess(drawing_a, "Sequential Average", "At least one channel must be active!");
    }
    break;
  case 12:  // two channel transformation;
    nactive=0; twochtrans->ch1=-1; twochtrans->ch2=-1;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	nactive++;
	if (twochtrans->ch1<0) twochtrans->ch1=i;
	else if (twochtrans->ch2<0) twochtrans->ch2=i;
      }
    }
    // exactly two channels are active;
    if (r.nvis>1 && nactive==2) {
      /* Generate Dialog Box */
      i=0;
      XtSetArg(args[i], XmNtitle, "Two channel transformations"); i++;
      XtSetArg(args[i], XmNautoUnmanage, False); i++;
      XtSetArg(args[i], XmNmessageString,
	       XmStringCreateLtoR("Select transformation:",
				  XmFONTLIST_DEFAULT_TAG)); i++;
      mbox=XmCreateMessageDialog(drawing_a, "twochtrans", args, i);
      // create rowcolumn widget;
      toggle_box=XmCreateRadioBox(mbox, "radiobox", NULL, 0);
      // create Radio buttons;
      twochtrans->rb[0]=XtVaCreateManagedWidget("a + b", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      twochtrans->rb[1]=XtVaCreateManagedWidget("a - b", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      twochtrans->rb[2]=XtVaCreateManagedWidget("b - a", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      twochtrans->rb[3]=XtVaCreateManagedWidget("a * b", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      twochtrans->rb[4]=XtVaCreateManagedWidget("a / b", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      twochtrans->rb[5]=XtVaCreateManagedWidget("b / a", xmToggleButtonGadgetClass, toggle_box,
						XmNset, False, NULL);
      XmToggleButtonGadgetSetState(twochtrans->rb[twochtrans->rb_selected],True,False);
      XtAddCallback(mbox, XmNokCallback, twochtrans_ok_cb, twochtrans);
      XtAddCallback(mbox, XmNcancelCallback, cancel_cb, NULL);
      XtAddCallback(mbox, XmNhelpCallback, twochtrans_help_cb, NULL);
      XtManageChild(toggle_box);
      XtManageChild(mbox);
      XtPopup(XtParent(mbox), XtGrabNone);
    }
    else { /* not exactly two channels are active */
      showmess(drawing_a, "Two channel transformations",
	       "Exactly two channels must be active!");
    }
    break;
  }
}


/*   .................   */
/*   c a l l b a c k s   */
/*   .................   */

void cancel_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  XtUnmanageChild(widget);
}

void derived_c_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  GC            gc;
  Display       *dpy=XtDisplay(drawing_a);
  Window        window=XtWindow(drawing_a);
  Dimension     wd, ht;
  Cursor        uhr, pfeil;
  Arg           args[5];
  Widget        mbox, *tb;
  short         dch[10]={0,0,0,0,0,0,0,0,0,0};
  int           i, nactive, chn, yunits, yoff, yd;
  int           x1, x2;
  unsigned long n, ni, na, ne, ndia, nsys, ndia_alt, nsys_alt;
  unsigned long nfirst, nlast;
  float         sum, sumt, sums, *data;
  float         data_min, data_max, data_mean, schwelle;
  float         dydt_min, dydt_max;

  extern void set_color(Widget widget, String fgcolor, String bgcolor, GC gc);
  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);
         void find_dia(float *data, unsigned long n1, unsigned long n2, float schwelle,
		       unsigned long *dia, unsigned long *sys);

  gc=XCreateGC(XtDisplay(drawing_a), RootWindowOfScreen(XtScreen(drawing_a)), 0, NULL);
  XtVaGetValues(drawing_a, XmNwidth, &wd, XmNheight, &ht, NULL);
  x1=LMAR+(int)((r.ts_1-r.t_1)*(float)(wd-LMAR)/(r.t_2-r.t_1));
  x2=LMAR+(int)((r.ts_2-r.t_1)*(float)(wd-LMAR)/(r.t_2-r.t_1));
  uhr  =XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil=XCreateFontCursor(XtDisplay(widget), XC_arrow);

  tb=(Widget *)client_data;

  nactive=0; yunits=0; chn=-1;
  for (i=0; i<r.n; i++) {
    if (r.view[i].is_active && r.view[i].is_visible) {
      nactive++;
      if (chn<0) chn=i;
    }
    if (r.view[i].is_visible) yunits+=r.ch[i].height+1;
  }

  if (r.nvis>0 && nactive==1) {
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    yoff=0;
    for (i=0; i<=chn; i++) {
      if(r.view[i].is_visible) {
	yd=(int)(((float)(ht*(r.ch[i].height+1)))/(float)yunits);
	yoff+=yd;
      }
    }
    set_color(drawing_a, "blue", NULL, gc);
    XDrawLine(dpy, window, gc, x1, yoff, x1, yoff-yd);
    XDrawLine(dpy, window, gc, x1, yoff-yd, x2, yoff-yd);
    XDrawLine(dpy, window, gc, x2, yoff-yd, x2, yoff);
    XDrawLine(dpy, window, gc, x2, yoff, x1, yoff);
    set_color(drawing_a, "black", NULL, gc);

    // fill data array with convolution;
    data=new float[r.ch[chn].n];
    na=(unsigned long)(r.ts_1*r.ch[chn].sr);
    if (na>=r.ch[chn].n) na=r.ch[chn].n-1;
    ne=(unsigned long)(r.ts_2*r.ch[chn].sr);
    if (ne>=r.ch[chn].n) ne=r.ch[chn].n-1;
    if (ne>na) {
      sumt=0; 
      for (ni=na; ni<=ne; ni++) 
	sumt+=r.ch[chn].data[ni] * r.ch[chn].data[ni];
      for (n=0; n<r.ch[chn].n-(ne-na+1); n++) {
	sums=0; sum=0;
	for (ni=na; ni<=ne; ni++) {
	  sums+=r.ch[chn].data[n+ni-na]  * r.ch[chn].data[n+ni-na];
	  sum +=r.ch[chn].data[n+ni-na]  * r.ch[chn].data[ni];
	}
	data[n]=sum/(sumt+sums);
      }
    }
    for (n=r.ch[chn].n-(ne-na+1); n<r.ch[chn].n; n++)
      data[n]=data[r.ch[chn].n-(ne-na+1)-1];

    // calculate average amplitude of data array and threashold;
    na=0; ne=r.ch[chn].n-2; n=na;
    find_dia(data, n, ne, 0, &ndia, &nsys);
    ndia_alt=ndia; nsys_alt=nsys; n=nsys;
    data_mean=data[nsys]-data[ndia]; ni=1;
    while (n<ne) {
      find_dia(data, n, ne, 0, &ndia, &nsys);
      if(nsys<ne){
	data_mean+=(data[nsys]-data[ndia]); ni++;
      }
      ndia_alt=ndia; nsys_alt=nsys;
      n=nsys;
    }
    schwelle=data_mean*0.5/ni;

    // create new channels;
    if (XmToggleButtonGadgetGetState(tb[0])) { /* frequency */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "frequency", "bpm");
      dch[0]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[1])) { /* maximum */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "maximum", r.ch[chn].unit);
      dch[1]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[2])) { /* mean */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "mean", r.ch[chn].unit);
      dch[2]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[3])) { /* minimum */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "minimum", r.ch[chn].unit);
      dch[3]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[4])) { /* amplitude */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "amplitude", r.ch[chn].unit);
      dch[4]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[5])) { /* dy/dt */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt", r.ch[chn].unit);
      for (n=0; n<r.ch[chn].n-1;  n++)
	r.ch[r.n-1].data[n]=(r.ch[chn].data[n+1]-r.ch[chn].data[n])*r.ch[chn].sr;
      r.ch[r.n-1].data[r.ch[chn].n-1]=r.ch[r.n-1].data[r.ch[chn].n-2];
      dch[5]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[6])) { /* dy/dtmax */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt_max", r.ch[chn].unit);
      dch[6]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[7])) { /* dy/dtmin */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt_min", r.ch[chn].unit);
      dch[7]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[8])) { /* trigger curve */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "trigger_curve", r.ch[chn].unit);
      for (n=0; n<r.ch[chn].n;  n++) r.ch[r.n-1].data[n]=data[n];
      dch[8]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[9])) { /* trigger value */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "trigger_value", r.ch[chn].unit);
      dch[9]=r.n-1;
    }

    // calculate derived channels;
    na=0; ne=r.ch[chn].n-2; n=na;
    find_dia(data, n, ne, schwelle, &ndia, &nsys);
    ndia_alt=ndia; nsys_alt=nsys; n=nsys; nfirst=nsys_alt;
    while (n<ne) {
      find_dia(data, n, ne, schwelle, &ndia, &nsys);
      if(nsys<ne){
	data_max=-1e30; data_min=1e30; data_mean=0;
	dydt_max=-1e30; dydt_min=1e30;
	for (ni=nsys_alt; ni<nsys; ni++) {
	  if (r.ch[chn].data[ni]>data_max) data_max=r.ch[chn].data[ni];
	  if (r.ch[chn].data[ni]<data_min) data_min=r.ch[chn].data[ni];
	  if ((r.ch[chn].data[ni+1]-r.ch[chn].data[ni])>dydt_max)
	    dydt_max=(r.ch[chn].data[ni+1]-r.ch[chn].data[ni]);
	  if ((r.ch[chn].data[ni+1]-r.ch[chn].data[ni])<dydt_min)
	    dydt_min=(r.ch[chn].data[ni+1]-r.ch[chn].data[ni]);
	  data_mean+=r.ch[chn].data[ni];
	}
	data_mean/=(nsys-nsys_alt);
	dydt_max*=r.ch[chn].sr;
	dydt_min*=r.ch[chn].sr;
	for (ni=nsys_alt; ni<nsys; ni++) {
	  if (XmToggleButtonGadgetGetState(tb[0]))
	    r.ch[dch[0]].data[ni]=r.ch[chn].sr/(nsys-nsys_alt);
	  if (XmToggleButtonGadgetGetState(tb[1]))
	    r.ch[dch[1]].data[ni]=data_max;
	  if (XmToggleButtonGadgetGetState(tb[2]))
	    r.ch[dch[2]].data[ni]=data_mean;
	  if (XmToggleButtonGadgetGetState(tb[3]))
	    r.ch[dch[3]].data[ni]=data_min;
	  if (XmToggleButtonGadgetGetState(tb[4]))
	    r.ch[dch[4]].data[ni]=data_max-data_min;
	  if (XmToggleButtonGadgetGetState(tb[6]))
	    r.ch[dch[6]].data[ni]=dydt_max;
	  if (XmToggleButtonGadgetGetState(tb[7]))
	    r.ch[dch[7]].data[ni]=dydt_min;
	  if (XmToggleButtonGadgetGetState(tb[9]))
	    r.ch[dch[9]].data[ni]=r.ch[chn].data[nsys_alt];
	}
	nlast=nsys;
      }
      ndia_alt=ndia; nsys_alt=nsys;
      n=nsys;
    }
    
    // fill first elements of new channels;
    for (n=0; n<nfirst; n++) {
      if (XmToggleButtonGadgetGetState(tb[0]))
	r.ch[dch[0]].data[n]=r.ch[dch[0]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[1]))
	r.ch[dch[1]].data[n]=r.ch[dch[1]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[2]))
	r.ch[dch[2]].data[n]=r.ch[dch[2]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[3]))
	r.ch[dch[3]].data[n]=r.ch[dch[3]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[4]))
	r.ch[dch[4]].data[n]=r.ch[dch[4]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[6]))
	r.ch[dch[6]].data[n]=r.ch[dch[6]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[7]))
	r.ch[dch[7]].data[n]=r.ch[dch[7]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[9]))
	r.ch[dch[9]].data[n]=r.ch[dch[9]].data[nfirst];
    }

    // fill last elements of new channels;
    for (n=nlast; n<r.ch[chn].n; n++) {
      if (XmToggleButtonGadgetGetState(tb[0]))
	r.ch[dch[0]].data[n]=r.ch[dch[0]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[1]))
	r.ch[dch[1]].data[n]=r.ch[dch[1]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[2]))
	r.ch[dch[2]].data[n]=r.ch[dch[2]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[3]))
	r.ch[dch[3]].data[n]=r.ch[dch[3]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[4]))
	r.ch[dch[4]].data[n]=r.ch[dch[4]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[6]))
	r.ch[dch[6]].data[n]=r.ch[dch[6]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[7]))
	r.ch[dch[7]].data[n]=r.ch[dch[7]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[9]))
	r.ch[dch[9]].data[n]=r.ch[dch[9]].data[nlast-1];
    }


    // set min and max for new channels and views;
    for (i=0; i<10; i++) {
      if (XmToggleButtonGadgetGetState(tb[i])) {
	r.view[dch[i]].is_active=0;
	for (n=0; n<r.ch[dch[i]].n; n++) {
	  if (r.ch[dch[i]].min > r.ch[dch[i]].data[n])
	    r.ch[dch[i]].min=r.ch[dch[i]].data[n];
	  if (r.ch[dch[i]].max < r.ch[dch[i]].data[n])
	    r.ch[dch[i]].max=r.ch[dch[i]].data[n];
	}
	na=(unsigned long)(r.t_1*r.ch[dch[i]].sr);
	if (na>=r.ch[dch[i]].n) na=r.ch[dch[i]].n-1;
	ne=(unsigned long)(r.t_2*r.ch[dch[i]].sr);
	if (ne>=r.ch[dch[i]].n) ne=r.ch[dch[i]].n-1;
	r.view[dch[i]].min=1e30; r.view[dch[i]].max=-1e30;
	for (n=na; n<=ne; n++) {
	  if (r.view[dch[i]].min > r.ch[dch[i]].data[n])
	    r.view[dch[i]].min=r.ch[dch[i]].data[n];
	  if (r.view[dch[i]].max < r.ch[dch[i]].data[n])
	    r.view[dch[i]].max=r.ch[dch[i]].data[n];
	}
      }
    }
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    delete [] data;
    XtUnmanageChild(widget);
    da_cb(NULL, NULL, NULL);
  }
  else {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("Exactly 1 channel must be active!",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNtitle, "Derived Channels"); i++;
    mbox=XmCreateMessageDialog(drawing_a, "Error !", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
    XtManageChild(mbox);
  }

  XFreeGC(dpy,gc);
  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
}

void derived_t_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Cursor        uhr, pfeil;
  Arg           args[5];
  Widget        mbox, *tb;
  char          *text;
  short         dch[8]={0,0,0,0,0,0,0,0};
  int           i, nactive, chn;
  unsigned long n, ni, na, ne, ndia, nsys, ndia_alt, nsys_alt;
  unsigned long nfirst, nlast;
  float         data_min, data_max, data_mean, schwelle;
  float         dydt_min, dydt_max;

  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);
         void find_dia(float *data, unsigned long n1, unsigned long n2, float schwelle,
		       unsigned long *dia, unsigned long *sys);
	 
  XmSelectionBoxCallbackStruct *cbs=(XmSelectionBoxCallbackStruct *) call_data;
  XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &text);
  schwelle=atof(text);
  XtFree(text);

  tb=(Widget *)client_data;

  uhr  =XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil=XCreateFontCursor(XtDisplay(widget), XC_arrow);

  nactive=0; chn=-1;
  for (i=0; i<r.n; i++) {
    if (r.view[i].is_active && r.view[i].is_visible) {
      nactive++;
      if (chn<0) chn=i;
    }
  }

  if (r.nvis>0 && nactive==1) {
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));

    // create new channels;
    if (XmToggleButtonGadgetGetState(tb[0])) { /* frequency */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "frequency", "bpm");
      dch[0]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[1])) { /* maximum */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "maximum", r.ch[chn].unit);
      dch[1]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[2])) { /* mean */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "mean", r.ch[chn].unit);
      dch[2]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[3])) { /* minimum */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "minimum", r.ch[chn].unit);
      dch[3]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[4])) { /* amplitude */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "minimum", r.ch[chn].unit);
      dch[4]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[5])) { /* dy/dt */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt", r.ch[chn].unit);
      for (n=0; n<r.ch[chn].n-1;  n++)
	r.ch[r.n-1].data[n]=(r.ch[chn].data[n+1]-r.ch[chn].data[n])*r.ch[chn].sr;
      r.ch[r.n-1].data[r.ch[chn].n-1]=r.ch[r.n-1].data[r.ch[chn].n-2];
      dch[5]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[6])) { /* dy/dtmax */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt_max", r.ch[chn].unit);
      dch[6]=r.n-1;
    }
    if (XmToggleButtonGadgetGetState(tb[7])) { /* dy/dtmin */
      r.AddChannel(r.ch[chn].n, r.ch[chn].sr, +1e30, -1e30, 0, 0, "dy/dt_min", r.ch[chn].unit);
      dch[7]=r.n-1;
    }

    // calculate derived channels;
    na=0; ne=r.ch[chn].n-2; n=na;
    find_dia(r.ch[chn].data, n, ne, schwelle, &ndia, &nsys);
    ndia_alt=ndia; nsys_alt=nsys; n=nsys; nfirst=nsys_alt;
    while (n<ne) {
      find_dia(r.ch[chn].data, n, ne, schwelle, &ndia, &nsys);
      if(nsys<ne){
	data_max=-1e30; data_min=1e30; data_mean=0;
	dydt_max=-1e30; dydt_min=1e30;
	for (ni=nsys_alt; ni<nsys; ni++) {
	  if (r.ch[chn].data[ni]>data_max) data_max=r.ch[chn].data[ni];
	  if (r.ch[chn].data[ni]<data_min) data_min=r.ch[chn].data[ni];
	  if ((r.ch[chn].data[ni+1]-r.ch[chn].data[ni])>dydt_max)
	    dydt_max=(r.ch[chn].data[ni+1]-r.ch[chn].data[ni]);
	  if ((r.ch[chn].data[ni+1]-r.ch[chn].data[ni])<dydt_min)
	    dydt_min=(r.ch[chn].data[ni+1]-r.ch[chn].data[ni]);
	  data_mean+=r.ch[chn].data[ni];
	}
	data_mean/=(nsys-nsys_alt);
	dydt_max*=r.ch[chn].sr;
	dydt_min*=r.ch[chn].sr;
	for (ni=nsys_alt; ni<nsys; ni++) {
	  if (XmToggleButtonGadgetGetState(tb[0]))
	    r.ch[dch[0]].data[ni]=r.ch[chn].sr/(nsys-nsys_alt);
	  if (XmToggleButtonGadgetGetState(tb[1]))
	    r.ch[dch[1]].data[ni]=data_max;
	  if (XmToggleButtonGadgetGetState(tb[2]))
	    r.ch[dch[2]].data[ni]=data_mean;
	  if (XmToggleButtonGadgetGetState(tb[3]))
	    r.ch[dch[3]].data[ni]=data_min;
	  if (XmToggleButtonGadgetGetState(tb[4]))
	    r.ch[dch[4]].data[ni]=data_max-data_min;
	  if (XmToggleButtonGadgetGetState(tb[6]))
	    r.ch[dch[6]].data[ni]=dydt_max;
	  if (XmToggleButtonGadgetGetState(tb[7]))
	    r.ch[dch[7]].data[ni]=dydt_min;
	}
	nlast=nsys;
      }
      ndia_alt=ndia; nsys_alt=nsys;
      n=nsys;
    }
    
    // fill first elements of new channels;
    for (n=0; n<nfirst; n++) {
      if (XmToggleButtonGadgetGetState(tb[0]))
	r.ch[dch[0]].data[n]=r.ch[dch[0]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[1]))
	r.ch[dch[1]].data[n]=r.ch[dch[1]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[2]))
	r.ch[dch[2]].data[n]=r.ch[dch[2]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[3]))
	r.ch[dch[3]].data[n]=r.ch[dch[3]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[4]))
	r.ch[dch[4]].data[n]=r.ch[dch[4]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[6]))
	r.ch[dch[6]].data[n]=r.ch[dch[6]].data[nfirst];
      if (XmToggleButtonGadgetGetState(tb[7]))
	r.ch[dch[7]].data[n]=r.ch[dch[7]].data[nfirst];
    }

    // fill last elements of new channels;
    for (n=nlast; n<r.ch[chn].n; n++) {
      if (XmToggleButtonGadgetGetState(tb[0]))
	r.ch[dch[0]].data[n]=r.ch[dch[0]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[1]))
	r.ch[dch[1]].data[n]=r.ch[dch[1]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[2]))
	r.ch[dch[2]].data[n]=r.ch[dch[2]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[3]))
	r.ch[dch[3]].data[n]=r.ch[dch[3]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[4]))
	r.ch[dch[4]].data[n]=r.ch[dch[4]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[6]))
	r.ch[dch[6]].data[n]=r.ch[dch[6]].data[nlast-1];
      if (XmToggleButtonGadgetGetState(tb[7]))
	r.ch[dch[7]].data[n]=r.ch[dch[7]].data[nlast-1];
    }


    // set min and max for new channels and views;
    for (i=0; i<8; i++) {
      if (XmToggleButtonGadgetGetState(tb[i])) {
	r.view[dch[i]].is_active=0;
	r.view[dch[i]].is_visible=1;
	for (n=0; n<r.ch[dch[i]].n; n++) {
	  if (r.ch[dch[i]].min > r.ch[dch[i]].data[n])
	    r.ch[dch[i]].min=r.ch[dch[i]].data[n];
	  if (r.ch[dch[i]].max < r.ch[dch[i]].data[n])
	    r.ch[dch[i]].max=r.ch[dch[i]].data[n];
	}
	na=(unsigned long)(r.t_1*r.ch[dch[i]].sr);
	if (na>=r.ch[dch[i]].n) na=r.ch[dch[i]].n-1;
	ne=(unsigned long)(r.t_2*r.ch[dch[i]].sr);
	if (ne>=r.ch[dch[i]].n) ne=r.ch[dch[i]].n-1;
	r.view[dch[i]].min=1e30; r.view[dch[i]].max=-1e30;
	for (n=na; n<=ne; n++) {
	  if (r.view[dch[i]].min > r.ch[dch[i]].data[n])
	    r.view[dch[i]].min=r.ch[dch[i]].data[n];
	  if (r.view[dch[i]].max < r.ch[dch[i]].data[n])
	    r.view[dch[i]].max=r.ch[dch[i]].data[n];
	}
      }
    }
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    XtUnmanageChild(widget);
    da_cb(NULL, NULL, NULL);
  }
  else {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("Exactly 1 channel must be active!",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNtitle, "Derived Channels"); i++;
    mbox=XmCreateMessageDialog(drawing_a, "Error !", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
    XtManageChild(mbox);
  }

  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
}

void filter_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  GC            gc;
  Display       *dpy=XtDisplay(drawing_a);
  Window        window=XtWindow(drawing_a);
  Dimension     ht, wd;
  Cursor        uhr, pfeil;
  Arg           args[5];
  Widget        mbox;
  filter_struct *filter;
  unsigned long fn, n, na, ne, n1, n2;
  int           i, j, nactive, chn;
  int           ford;
  int           yoff, yd;
  float         ave, adev, sdev, var, skew, curt;
  float         yf, ymax, ymin, yunits;
  float         fc, wg;
  float         *f,*h;
  short         lh;

  extern void set_color(Widget widget, String fgcolor, String bgcolor, GC gc);
  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  extern void showmess(Widget widget, char *title, char *text);

  gc=XCreateGC(XtDisplay(drawing_a), RootWindowOfScreen(XtScreen(drawing_a)), 0, NULL);
  XtVaGetValues(drawing_a, XmNwidth, &wd, XmNheight, &ht, NULL);
  filter = (filter_struct *) client_data;
  uhr    = XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil  = XCreateFontCursor(XtDisplay(widget), XC_arrow);

  /* testing whether at least one channel is active */
  nactive=0; chn=-1, yunits=0;;
  for (i=0; i<r.n; i++) {
    if (r.view[i].is_active && r.view[i].is_visible) {
      nactive++;
      if (chn<0) chn=i;
    }
    if (r.view[i].is_visible)
      yunits+=r.ch[i].height+1;
  }

  /* if at least one channel is active do something */
  if (r.nvis>0 && nactive>0) {
    /* Parameter Definition */
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    fc   = filter->freq;
    ford = filter->order;
    lh   = filter->type;

    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	/* determine 2^n for FFT */
	if (floor(log(r.ch[i].n)/log(2))==log(r.ch[i].n)/log(2))
	  fn=(unsigned long)pow(2,floor(log(r.ch[i].n)/log(2))); 
	else  
	  fn=(unsigned long)pow(2,floor(log(r.ch[i].n)/log(2))+1); 
	/* Dataarray Initialization */
	f=new float[2*fn*sizeof(float)];
	h=new float[2*fn*sizeof(float)];

	/* Filter definieren */
	wg=fc*fn/r.ch[i].sr;
	for (n=0; n<=fn/2; n++) {
	  if (lh==0) h[2*n]=sqrt(1/(1+pow(n/wg,2*ford)));
	  else       h[2*n]=sqrt(1/(1+pow(wg/(n+0.0001),2*ford)));
	  h[2*n+1]=0;
	} /* endfor */
	for (n=1; n<fn/2; n++) {
	  h[fn+n*2]=h[fn-n*2];
	  h[fn+n*2+1]=0;
	} /* endfor */

	/* mean for zero padding */
	moment(r.ch[i].data-1,r.ch[i].n,&ave,&adev,&sdev,&var,&skew,&curt);

	/* Daten kopieren */
	for (n=0; n<r.ch[i].n; n++) {
	  f[2*n]=r.ch[i].data[n]-ave;
	  f[2*n+1]=0;
	} /* endfor */
	for (n=r.ch[i].n; n<fn; n++) {
	  f[2*n]=0;
	  f[2*n+1]=0;
	} /* endfor */

	/* filter */
	four1(f-1,fn,1);
	for (n=0; n<fn; n++) { 
	  f[2*n]   = f[2*n]  * h[2*n];
	  f[2*n+1] = f[2*n+1]* h[2*n];
	} /* endfor */  
	four1(f-1,fn,-1);

	/* gefilterte Daten kopieren */
	for (n=0; n<r.ch[i].n; n++) {
	  if (lh==0) r.ch[i].data[n]=f[n*2]/fn+ave;
	  else       r.ch[i].data[n]=f[n*2]/fn;
	}

	/* Show filtered curve */
	yoff=0;
	for (j=0; j<=i; j++) {
	  if (r.view[j].is_visible) {
	    yd=(int)(((float)(ht*(r.ch[j].height+1)))/(float)yunits);
	    yoff+=yd;
	  }
	}
	yf=(float)yd/(r.view[i].max-r.view[i].min);
	set_color(drawing_a, "red", NULL, gc);
	na=(unsigned long)(r.t_1*r.ch[i].sr);
	if (na>=r.ch[i].n) na=r.ch[i].n-1;
	n2=na+1;
	for (j=LMAR; j<=wd; j++) {
	  n1=n2;
	  if (n1<0) n1=0;
	  n2=(unsigned long)((r.t_1*r.ch[i].sr)+
			     (r.t_2-r.t_1)*r.ch[i].sr*(j+1-LMAR)/(wd-LMAR));
	  if (n2>=r.ch[i].n) n2=r.ch[i].n-1;
	  if (n1>n2) n2=n1;
	  ymax=r.ch[i].data[n1];
	  ymin=r.ch[i].data[n1];
	  for (n=n1; n<=n2; n++) {
	    if (r.ch[i].data[n]>ymax) ymax=r.ch[i].data[n];
	    if (r.ch[i].data[n]<ymin) ymin=r.ch[i].data[n];
	  }
	  if (lh==1) {
	    ymin+=ave;
	    ymax+=ave;
	  }
	  XDrawLine(dpy,window,gc,j,(int)(yoff-yf*(ymin-r.view[i].min)),
		    j,(int)(yoff-yf*(ymax-r.view[i].min)));
	}
	set_color(drawing_a, "black", NULL, gc);

	/* Min und Max fuer Kanal setzen */
	na=(unsigned long)(r.t_1*r.ch[i].sr);
	if (na>=r.ch[i].n) na=r.ch[i].n-1;
	ne=(unsigned long)(r.t_2*r.ch[i].sr);
	if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	r.view[i].min=1e30; r.view[i].max=-1e30;
	for (n=na; n<=ne; n++) {
	  if (r.view[i].min > r.ch[i].data[n])
	    r.view[i].min=r.ch[i].data[n];
	  if (r.view[i].max < r.ch[i].data[n])
	    r.view[i].max=r.ch[i].data[n];
	}
	r.ch[i].min=1e30;   r.ch[i].max=-1e30;
	for (n=0; n<r.ch[i].n; n++) {
	  if (r.ch[i].data[n]<r.ch[i].min) r.ch[i].min=r.ch[i].data[n];
	  if (r.ch[i].data[n]>r.ch[i].max) r.ch[i].max=r.ch[i].data[n];
	} /* endfor */
      } /* endif */
    } /* endfor */

    /* done with filtering */
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    delete[]f; delete[]h;
  }
  else {
    showmess(drawing_a, "Butterworth Filter", "At least 1 channel must be active!");
  }

  XFreeGC(dpy,gc);
  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
  XtUnmanageChild(widget);
}

void linear_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  linear_struct *linear;
  int           i, nactive;
  unsigned long n;
  float         fdummy;
  Cursor        uhr, pfeil;
  Widget        mbox;
  Arg           args[5];

  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);

  uhr    = XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil  = XCreateFontCursor(XtDisplay(widget), XC_arrow);

  nactive=0;
  for (i=0; i<r.n; i++) {
    if (r.view[i].is_active && r.view[i].is_visible) {
      nactive++;
    }
  }

  /* if at least one channel is active do something */
  if (r.nvis>0 && nactive>0) {
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    /* get linear struct */
    linear=(linear_struct *) client_data;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	for (n=0; n<r.ch[i].n; n++) {
	  r.ch[i].data[n]=r.ch[i].data[n]*linear->m + linear->b;
	} /* endfor */
	/* Min und Max fuer Kanal setzen */
	r.view[i].min=r.view[i].min*linear->m + linear->b;
	r.view[i].max=r.view[i].max*linear->m + linear->b;
	if (r.view[i].min > r.view[i].max) {
	  fdummy=r.view[i].min;
	  r.view[i].min=r.view[i].max;
	  r.view[i].max=fdummy;
	}
	r.ch[i].min=r.ch[i].min*linear->m + linear->b;
	r.ch[i].max=r.ch[i].max*linear->m + linear->b;
	if (r.ch[i].min > r.ch[i].max) {
	  fdummy=r.ch[i].min;
	  r.ch[i].min=r.ch[i].max;
	  r.ch[i].max=fdummy;
	}
      } /* endif */
    } /* endfor */
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    XtUnmanageChild(widget);
    da_cb(NULL, NULL, NULL);
  } /* endif */
  else {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("At least one channel must be active!",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNtitle, "Linear Transformation"); i++;
    mbox=XmCreateMessageDialog(drawing_a, "Error !", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
    XtManageChild(mbox);
  } /* endelse */
  
  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
}

void interpol_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int           i, nactive;
  unsigned long n, ndata;
  float         *interpol_sr;
  float         *xdata, *y2, *ydata;
  float         dx, xpos, wert;
  Cursor        uhr, pfeil;
  Widget        mbox;
  Arg           args[5];

  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);

  uhr    = XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil  = XCreateFontCursor(XtDisplay(widget), XC_arrow);

  /* how many channels are active ? */
  nactive=0;
  for (i=0; i<r.n; i++) 
    if (r.view[i].is_active && r.view[i].is_visible) nactive++;

  /* if at least one channel is active do something */
  if (r.nvis>0 && nactive>0) {
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    /* get interpolation sampling rate */
    interpol_sr=(float *) client_data;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	// calculate spline interpolation;
	xdata=new float[r.ch[i].n];
	y2   =new float[r.ch[i].n];
	for (n=0; n<r.ch[i].n; n++) xdata[n]=(float)n/r.ch[i].sr;
	spline(xdata-1, r.ch[i].data-1, r.ch[i].n, 1e30, 1e30, y2-1);
	xpos=0.0; dx=1.0/(*interpol_sr);
	ndata=(unsigned long)((float)((r.ch[i].n-1)/r.ch[i].sr)*(*interpol_sr));
	ydata=new float[ndata];
	for (n=0; n<ndata; n++) {
	  splint(xdata-1, r.ch[i].data-1, y2-1, r.ch[i].n, xpos, &wert);
	  ydata[n]=wert;
	  xpos+=dx;
	}
	r.ch[i].data=(float *)realloc(r.ch[i].data, ndata*sizeof(float));
	for (n=0; n<ndata; n++) r.ch[i].data[n]=ydata[n];
	r.ch[i].sr=*interpol_sr;
	r.ch[i].n=ndata;
	delete[]ydata;
	delete[]y2;
	delete[]xdata;
      } /* endif */
    } /* endfor */
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    XtUnmanageChild(widget);
    da_cb(NULL, NULL, NULL);
  } /* endif */
  else {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("At least one channel must be active!",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNtitle, "Spline Interpolation"); i++;
    mbox=XmCreateMessageDialog(drawing_a, "Error !", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
    XtManageChild(mbox);
  } /* endelse */
  
  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
}

void derived_c_help_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int    i;
  Arg    args[5];
  Widget mbox;

  i=0;
  XtSetArg(args[i], XmNmessageString,
	   XmStringCreateLtoR("Make exactly one channel active (3rd mouse button). \nMark a template that represents one wavelength. \nSelect the parameters to be calculated.",
	   XmFONTLIST_DEFAULT_TAG)); i++;
  XtSetArg(args[i], XmNtitle, "Derived Channels"); i++;
  mbox=XmCreateMessageDialog(drawing_a, "Help !", args, i);
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
  XtManageChild(mbox);
}

void derived_t_help_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int    i;
  Arg    args[5];
  Widget mbox;

  i=0;
  XtSetArg(args[i], XmNmessageString,
	   XmStringCreateLtoR("Make exactly one channel active (3rd mouse button). \nType in a threshold value for triggering an amplitude. \nSelect the parameters to be calculated.",
	   XmFONTLIST_DEFAULT_TAG)); i++;
  XtSetArg(args[i], XmNtitle, "Derived Channels"); i++;
  mbox=XmCreateMessageDialog(drawing_a, "Help !", args, i);
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
  XtManageChild(mbox);
}

void filter_help_cb(Widget widget, XtPointer client_data, XtPointer call_data){
  int    i;
  Arg    args[5];
  Widget mbox;

  i=0;
  XtSetArg(args[i], XmNmessageString,
	   XmStringCreateLtoR("Make exactly one channel active (3rd mouse button). \nSelect either low or high pass filter. \nChoose a filter order.\nType in a corner frequency.",
	   XmFONTLIST_DEFAULT_TAG)); i++;
  XtSetArg(args[i], XmNtitle, "Butterworth Filter"); i++;
  mbox=XmCreateMessageDialog(drawing_a, "Help !", args, i);
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
  XtManageChild(mbox);
}

void linear_help_cb(Widget widget, XtPointer client_data, XtPointer call_data){
  int    i;
  Arg    args[5];
  Widget mbox;

  i=0;
  XtSetArg(args[i], XmNmessageString,
	   XmStringCreateLtoR("Activate at least one channel (3rd mouse button). \nType in parameters m and b for linear transformation.",
	   XmFONTLIST_DEFAULT_TAG)); i++;
  XtSetArg(args[i], XmNtitle, "Linear Transformation"); i++;
  mbox=XmCreateMessageDialog(drawing_a, "Help !", args, i);
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
  XtManageChild(mbox);
}

void interpol_help_cb(Widget widget, XtPointer client_data, XtPointer call_data){
  int    i;
  Arg    args[5];
  Widget mbox;

  i=0;
  XtSetArg(args[i], XmNmessageString,
	   XmStringCreateLtoR("Activate at least one channel (3rd mouse button).\nType in the new sampling rate.\nIf sampling rate is reduced, there is no low-pass filtering.\nThis may result in aliasiing!",
	   XmFONTLIST_DEFAULT_TAG)); i++;
  XtSetArg(args[i], XmNtitle, "Spline Interpolation"); i++;
  mbox=XmCreateMessageDialog(drawing_a, "Help !", args, i);
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
  XtManageChild(mbox);
}

void baro_help_cb(Widget widget, XtPointer client_data, XtPointer call_data){

  extern void showmess(Widget widget, char *title, char *text);

  showmess(widget, "Baroreflex",
	   "1. Activate the BP channel\n2. Enter a channel number for the y-axis variable\n   (visible channels are counted from top to bottom)\n   (0: heart rate calculated from BP)\n3. Enter threshold for detection of BP-amplitude \n4. Enter a minimal regression coefficient\n   for inclusion of sequence.");
}


/* ............................. */
/* ...   various functions   ... */
/* ............................. */

/* function find_dia */
void find_dia(float *data, unsigned long n1, unsigned long n2, float schwelle,
	      unsigned long *dia, unsigned long *sys) {
  unsigned long n;

  n=n1+1;
  do {
    /* suche Diastole */
    while((data[n]>=data[n-1] || data[n]>data[n+1]) && (n<n2))
      n++;
    *dia=n;
    /* suche Systole */
    while((data[n]<data[n-1] || data[n]<=data[n+1]) && (n<n2))
      n++;
    *sys=n;
  }
  while ((data[*sys]-data[*dia])<=schwelle && n<n2);

  if (n>=n2) {*dia=n2; *sys=n2;};
  return;
}


/* function for low or high pass toggles */
void filter_tg_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  filter_struct *filter;
  int           lowhigh;

  XmToggleButtonCallbackStruct *state = (XmToggleButtonCallbackStruct *) call_data;
  XtVaGetValues(widget, XmNuserData, &filter, NULL);
  lowhigh=(int) client_data;

  if (lowhigh==0 && state->set) filter->type=0;
  if (lowhigh==1 && state->set) filter->type=1;
}

/* function for filter order */
void filter_ord_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  filter_struct *filter;
  int           order;

  filter=(filter_struct *) client_data;
  XmScaleGetValue(widget, &order);
  filter->order=order;
}

/* function for filter frequency */
void filter_freq_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  filter_struct *filter;
  char          *text;

  filter=(filter_struct *) client_data;
  text=XmTextFieldGetString(widget);
  
  if (atof(text)>0)
    filter->freq=atof(text);

  XtFree(text);
}

/* function for linear parameter m */
void linear_pm_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  linear_struct *linear;
  char          *text;

  linear=(linear_struct *) client_data;
  text=XmTextFieldGetString(widget);

  linear->m=atof(text);
  
  XtFree(text);
}

/* function for linear paramter b */
void linear_pb_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  linear_struct *linear;
  char          *text;

  linear=(linear_struct *) client_data;
  text=XmTextFieldGetString(widget);

  linear->b=atof(text);
  
  XtFree(text);
}

/* function for interpolation sampling rate */
void interpol_sr_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  char       *text;
  float      *interpol_sr;

  interpol_sr=(float *) client_data;
  text=XmTextFieldGetString(widget);
  
  if (atof(text)>0)
    *interpol_sr=atof(text);

  XtFree(text);
}

/* function for selection of SYS, Mean, DIA for Baroreflex calculation */
void baro_tb_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  baro_struct *baro;
  int         i;

  XtVaGetValues(widget, XmNuserData, &i, NULL);
  baro=(baro_struct *) client_data;
  baro->smd=(short)i;
}

/* function for baro channel */
void baro_ch_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  baro_struct *baro;
  char        *text;

  baro=(baro_struct *) client_data;
  text=XmTextFieldGetString(widget);

  baro->ch=atoi(text);
  
  XtFree(text);
}

/* function for baro threshold */
void baro_thre_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  baro_struct  *baro;
  char         *text;

  baro=(baro_struct *) client_data;
  text=XmTextFieldGetString(widget);

  baro->thre=atof(text);
  
  XtFree(text);
}

/* function for baro regression coefficient */
void baro_reg_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  baro_struct *baro;
  char        *text;

  baro=(baro_struct *) client_data;
  text=XmTextFieldGetString(widget);

  baro->reg=atof(text);
  
  XtFree(text);
}

/* function to set number of values to average */
void mavg_n_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  char          *text, **rc;
  unsigned long n;
  float         tau;
  mavg_struct   *mavg;

  mavg=(mavg_struct *) client_data;
  text=XmTextFieldGetString(widget);
  
  tau=atof(text);
  if (tau>0.0) {
    mavg->tau=tau;
    mavg->is_error=0;
  }
  else mavg->is_error=1;
  XtFree(text);
}

/* function to set if sampling rate should be reduced */
void mavg_tg_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  mavg_struct *mavg;
  mavg=(mavg_struct *) client_data;
  if (XmToggleButtonGadgetGetState(widget))
    mavg->reduce=1;
  else
    mavg->reduce=0;
}

/* OK function for moving average */
void mavg_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  mavg_struct   *mavg;
  GC            gc;
  Display       *dpy=XtDisplay(drawing_a);
  Window        window=XtWindow(drawing_a);
  Dimension     ht, wd;
  Cursor        uhr, pfeil;
  int           i, j, nactive;
  unsigned long nmavg, n, ndata, nsum, na, ne, ni, n1, n2;
  float         *data, sum;
  int           yoff, yd;
  float         yf, ymax, ymin, yunits;

  extern void showmess(Widget widget, char *title, char *text);
  extern void set_color(Widget widget, String fgcolor, String bgcolor, GC gc);

  mavg=(mavg_struct *) client_data;
  gc=XCreateGC(XtDisplay(drawing_a), RootWindowOfScreen(XtScreen(drawing_a)), 0, NULL);
  XtVaGetValues(drawing_a, XmNwidth, &wd, XmNheight, &ht, NULL);

  if (!mavg->is_error) {
    uhr    = XCreateFontCursor(XtDisplay(widget),XC_watch);
    pfeil  = XCreateFontCursor(XtDisplay(widget), XC_arrow);

    /* how many channels are active ? */
    nactive=0; yunits=0;
    for (i=0; i<r.n; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) nactive++;
      if (r.view[i].is_visible) yunits+=r.ch[i].height+1;
    }

    /* if at least one channel is active do something */
    if (r.nvis>0 && nactive>0) {
      XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
      XFlush(XtDisplay(widget));
      for (i=0; i<r.n; i++) {
	if (r.view[i].is_active && r.view[i].is_visible) {
	  nmavg=(unsigned long)((r.ch[i].sr*mavg->tau)+0.5);
	  /* do moving average and reduce */
	  if (nmavg>1 && mavg->reduce) {
	    data=new float[2*((r.ch[i].n/nmavg)+1)]; /* make it twice as large to be save */
	    ndata=0;
	    for (n=0; n<r.ch[i].n-nmavg; n+=nmavg) {
	      nsum=0; sum=0.0;
	      for (ni=0; ni<nmavg; ni++) {
		sum+=r.ch[i].data[n+ni]; nsum++;
	      }
	      data[ndata]=sum/nsum; ndata++;
	    }
	    r.ch[i].data=(float *)realloc(r.ch[i].data, ndata*sizeof(float));
	    for (n=0; n<ndata; n++) r.ch[i].data[n]=data[n];
	    r.ch[i].sr=r.ch[i].sr/nmavg;
	    r.ch[i].n=ndata;
	    delete [] data;
	  }
	  /* do moving average and do not reduce */
	  if (nmavg>1 && !mavg->reduce) {
	    data=new float[r.ch[i].n];
	    for (n=0; n<r.ch[i].n; n++) {
	      nsum=0; sum=0;
	      for (ni=0; ni<=nmavg; ni++) {
		if (n+ni>=nmavg/2 && n+ni-nmavg/2<r.ch[i].n) {
		  sum+=r.ch[i].data[n+ni-nmavg/2]; nsum++;
		}
	      }
	      data[n]=sum/nsum;
	    }
	    for (n=0; n<r.ch[i].n; n++) r.ch[i].data[n]=data[n];
	    delete [] data;
	  }
	  /* Show filtered curve */
	  if (nmavg>1) {
	    yoff=0;
	    for (j=0; j<=i; j++) {
	      if (r.view[j].is_visible) {
		yd=(int)(((float)(ht*(r.ch[j].height+1)))/(float)yunits);
		yoff+=yd;
	      }
	    }
	    yf=(float)yd/(r.view[i].max-r.view[i].min);
	    set_color(drawing_a, "red", NULL, gc);
	    na=(unsigned long)(r.t_1*r.ch[i].sr);
	    if (na>=r.ch[i].n) na=r.ch[i].n-1;
	    n2=na+1;
	    for (j=LMAR; j<=wd; j++) {
	      n1=n2;
	      if (n1<0) n1=0;
	      n2=(unsigned long)((r.t_1*r.ch[i].sr)+
				 (r.t_2-r.t_1)*r.ch[i].sr*(j+1-LMAR)/(wd-LMAR));
	      if (n2>=r.ch[i].n) n2=r.ch[i].n-1;
	      if (n1>n2) n2=n1;
	      ymax=r.ch[i].data[n1];
	      ymin=r.ch[i].data[n1];
	      for (n=n1; n<=n2; n++) {
		if (r.ch[i].data[n]>ymax) ymax=r.ch[i].data[n];
		if (r.ch[i].data[n]<ymin) ymin=r.ch[i].data[n];
	      }
	      XDrawLine(dpy,window,gc,j,(int)(yoff-yf*(ymin-r.view[i].min)),
			j,(int)(yoff-yf*(ymax-r.view[i].min)));
	    }
	    set_color(drawing_a, "black", NULL, gc);

	    /* Min und Max fuer Kanal setzen */
	    na=(unsigned long)(r.t_1*r.ch[i].sr);
	    if (na>=r.ch[i].n) na=r.ch[i].n-1;
	    ne=(unsigned long)(r.t_2*r.ch[i].sr);
	    if (ne>=r.ch[i].n) ne=r.ch[i].n-1;
	    r.view[i].min=1e30; r.view[i].max=-1e30;
	    for (n=na; n<=ne; n++) {
	      if (r.view[i].min > r.ch[i].data[n])
		r.view[i].min=r.ch[i].data[n];
	      if (r.view[i].max < r.ch[i].data[n])
		r.view[i].max=r.ch[i].data[n];
	    }
	    r.ch[i].min=1e30;   r.ch[i].max=-1e30;
	    for (n=0; n<r.ch[i].n; n++) {
	      if (r.ch[i].data[n]<r.ch[i].min) r.ch[i].min=r.ch[i].data[n];
	      if (r.ch[i].data[n]>r.ch[i].max) r.ch[i].max=r.ch[i].data[n];
	    } /* endfor */
	  } /* endif */
	} /* endfor */
      } /* endif */

      XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    }
    XFreeCursor(XtDisplay(widget), pfeil);
    XFreeCursor(XtDisplay(widget), uhr);
    XtUnmanageChild(widget);
  }
  else {
    showmess(widget, "Error", "Enter a correct time for moving average.");
  }
  XFreeGC(dpy,gc);
}

/* help function for moving average */
void mavg_help_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
extern void showmess(Widget widget, char *title, char *text);
showmess(widget, "Moving Average",
  "Enter time in seconds over that the\ntime series should be averaged.\nIf reduce sampling rate is activated\nThe new time series will contain only\none value for each moving average time segment,\ni.e. the new sampling rate is 1/time Hz.");
}


/* Functions for sequential average */
void seqavg_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  float *wert;
  char  *text;
  wert=(float *) client_data;
  text=XmTextFieldGetString(widget);
  *wert=atof(text);
  XtFree(text);
}

void seqavg_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  savg_struct   *savg;
  int           i, nactive;
  short         nch;
  unsigned long n, ni, nseq, nvor, nnach, nl;
  unsigned long ndia, ndia_alt, nsys, nsys_alt,  na, ne;
  float         fdummy;
  Cursor        uhr, pfeil;
  Widget        mbox;
  Arg           args[5];

  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);
         void find_dia(float *data, unsigned long n1, unsigned long n2,
		       float schwelle,
		       unsigned long *dia, unsigned long *sys);

  uhr    = XCreateFontCursor(XtDisplay(widget),XC_watch);
  pfeil  = XCreateFontCursor(XtDisplay(widget), XC_arrow);

  nactive=0;
  for (i=0; i<r.n; i++) {
    if (r.view[i].is_active && r.view[i].is_visible) {
      nactive++;
    }
  }
  /* if at least one channel is active do something */
  if (r.nvis>0 && nactive>0) {
    XDefineCursor(XtDisplay(widget),XtWindow(widget),uhr);
    XFlush(XtDisplay(widget));
    /* get length of sequence and schwelle */
    savg=(savg_struct *) client_data;
    nch=r.n;
    for (i=0; i<(int)nch; i++) {
      if (r.view[i].is_active && r.view[i].is_visible) {
	/* determine number of values for sequence */
	nvor =(unsigned long)(savg->vor *r.ch[i].sr);
	nnach=(unsigned long)(savg->nach*r.ch[i].sr);
	nseq = nvor + nnach;
	if (nseq>0) {
	  /* add new channel */
	  r.AddChannel(nseq, r.ch[i].sr, 1e30, -1e30,
		       r.ch[i].height, r.ch[i].color,
		       r.ch[i].name, r.ch[i].unit);
	  /* initiate new data field */
	  for (n=0; n<nseq; n++) r.ch[r.n-1].data[n]=0;
	  /* find first local minimum */
	  na=0; ne=r.ch[i].n-2;
	  find_dia(r.ch[i].data, na, ne, savg->schwelle, &ndia, &nsys);
	  ndia_alt=ndia; nsys_alt=nsys;
	  /* define start point for new dia search */
	  if (nsys>(ndia+nnach)) na=nsys;
	  else na=ndia+nnach;
	  nl=0; /* counter for number of sequences */
	  while (na<ne) {
	    find_dia(r.ch[i].data, na, ne, savg->schwelle, &ndia, &nsys);
	    if ((ndia+nnach<ne) && (ndia-nvor>=0)) {
	      for (ni=ndia-nvor; ni<ndia+nnach; ni++) {
		r.ch[r.n-1].data[ni-ndia+nvor]+=(r.ch[i].data[ni]);
	      }
	      nl++;
	    }
	    ndia_alt=ndia; nsys_alt=nsys;
	    /* define start point for new dia search */
	    if (nsys>(ndia+nnach)) na=nsys;
	    else na=ndia+nnach;
	  }
	  /* divide by number of sequences for mean */
	  for (ni=0; ni<nseq; ni++) {
	    r.ch[r.n-1].data[ni]/=nl;
	    /* Min und Max fuer neuen Kanal setzen */
	    if (r.ch[r.n-1].data[ni]<r.ch[r.n-1].min)
	      r.ch[r.n-1].min=r.ch[r.n-1].data[ni];
	    if (r.ch[r.n-1].data[ni]>r.ch[r.n-1].max)
	      r.ch[r.n-1].max=r.ch[r.n-1].data[ni];
	  }
	  /* set min and max for view */
	  r.view[r.n-1].min = r.ch[r.n-1].min;
	  r.view[r.n-1].max = r.ch[r.n-1].max;
	} /* endif >0 */
      } /* endif */
    } /* endfor */
    XDefineCursor(XtDisplay(widget),XtWindow(widget),pfeil);
    XtUnmanageChild(widget);
    da_cb(NULL, NULL, NULL);
  } /* endif */
  else {
    i=0;
    XtSetArg(args[i], XmNmessageString, XmStringCreateLtoR("At least one channel must be active!",XmFONTLIST_DEFAULT_TAG)); i++;
    XtSetArg(args[i], XmNtitle, "Sequential Average"); i++;
    mbox=XmCreateMessageDialog(drawing_a, "Error !", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(mbox, XmDIALOG_CANCEL_BUTTON));
    XtManageChild(mbox);
  } /* endelse */
  
  XFreeCursor(XtDisplay(widget), pfeil);
  XFreeCursor(XtDisplay(widget), uhr);
}

void seqavg_help_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
extern void showmess(Widget widget, char *title, char *text);
showmess(widget, "Sequential  Average",
  "Averages sequences of the time series\naround a pattern dected by a threshold.\nEnter time in seconds before and after\nthe pattern and the threshold.");
}

void twochtrans_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  twochtrans_struct *tct;
  unsigned long     n;
  short             is_valid, ndivz;
  char              text[256];

  extern void showmess(Widget widget, char *title, char *text);
  extern void da_cb(Widget widget, XtPointer client_data, XtPointer call_data);

  /* get data */
  tct=(twochtrans_struct *) client_data;

  /* sanity check */
  is_valid=1;
  /* sampling rate */
  if (r.ch[tct->ch1].sr != r.ch[tct->ch2].sr) is_valid=0;
  /* number of data points */
  if (r.ch[tct->ch1].n != r.ch[tct->ch2].n) is_valid=0;

  /* everything is ok, lets do it */
  if (is_valid) {
    /* add new channel */
    r.AddChannel(r.ch[tct->ch1].n, r.ch[tct->ch1].sr, 1e30, -1e30,
		 0, 1, "twochtrans", "tct_units");
    ndivz=0;
    for (n=0; n<r.ch[tct->ch1].n; n++) {
      if (XmToggleButtonGadgetGetState(tct->rb[0])) { /* a+b */
	r.ch[r.n-1].data[n]=r.ch[tct->ch1].data[n]+r.ch[tct->ch2].data[n];
	tct->rb_selected=0;
      }
      if (XmToggleButtonGadgetGetState(tct->rb[1])) { /* a-b */
	r.ch[r.n-1].data[n]=r.ch[tct->ch1].data[n]-r.ch[tct->ch2].data[n];
	tct->rb_selected=1;
      }
      if (XmToggleButtonGadgetGetState(tct->rb[2])) { /* b-a */
	r.ch[r.n-1].data[n]=r.ch[tct->ch2].data[n]-r.ch[tct->ch1].data[n];
	tct->rb_selected=2;
      }
      if (XmToggleButtonGadgetGetState(tct->rb[3])) { /* a*b */
	r.ch[r.n-1].data[n]=r.ch[tct->ch1].data[n]*r.ch[tct->ch2].data[n];
	tct->rb_selected=3;
      }
      if (XmToggleButtonGadgetGetState(tct->rb[4])) { /* a/b */
	if (r.ch[tct->ch2].data[n]!=0) 
	  r.ch[r.n-1].data[n]=r.ch[tct->ch1].data[n]/r.ch[tct->ch2].data[n];
	else {
	  r.ch[r.n-1].data[n]=0;
	  ndivz++;
	}
	tct->rb_selected=4;
      }
      if (XmToggleButtonGadgetGetState(tct->rb[5])) { /* b/a */
	if (r.ch[tct->ch1].data[n]!=0) 
	  r.ch[r.n-1].data[n]=r.ch[tct->ch2].data[n]/r.ch[tct->ch1].data[n];
	else {
	  r.ch[r.n-1].data[n]=0;
	  ndivz++;
	}
	tct->rb_selected=5;
      }
      /* Min und Max fuer neuen Kanal setzen */
      if (r.ch[r.n-1].data[n]<r.ch[r.n-1].min)
	r.ch[r.n-1].min=r.ch[r.n-1].data[n];
      if (r.ch[r.n-1].data[n]>r.ch[r.n-1].max)
	r.ch[r.n-1].max=r.ch[r.n-1].data[n];
    }
    if (ndivz>0) {
      sprintf(text,
	      "There were %d divisions by 0!\nValues substituted by 0",
	      ndivz);
      showmess(drawing_a, "Two Channel Transformations", text);
    }
    /* set min and max for view */
    r.view[r.n-1].min = r.ch[r.n-1].min;
    r.view[r.n-1].max = r.ch[r.n-1].max;
    da_cb(NULL, NULL, NULL);
  }
  /* parameters do not match */
  else {
    showmess(drawing_a, "Two Channel Transformations", "The two channels do not match!");
  }

  XtUnmanageChild(widget);
}

void twochtrans_help_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
extern void showmess(Widget widget, char *title, char *text);
showmess(widget, "Two Channel Transformations",
  "Two channels must be selected.\na: upper channel\nb: lower channel\nSelect the transformation to be applied to the two channels.\nMake sure channel b is not 0 if using a / b !");
}
